@@ -301,6 +301,7 @@ void dav1d_create_lf_mask_intra(Av1Filter *const lflvl,
     const int bh4 = imin(ih - by, b_dim[1]);
     const int bx4 = bx & 31;
     const int by4 = by & 31;
+    assert(bw4 >= 0 && bh4 >= 0);
 
     if (bw4 && bh4) {
         uint8_t (*level_cache_ptr)[4] = level_cache + by * b4_stride + bx;
@@ -323,6 +324,7 @@ void dav1d_create_lf_mask_intra(Av1Filter *const lflvl,
                           (b_dim[0] + ss_hor) >> ss_hor);
     const int cbh4 = imin(((ih + ss_ver) >> ss_ver) - (by >> ss_ver),
                           (b_dim[1] + ss_ver) >> ss_ver);
+    assert(cbw4 >= 0 && cbh4 >= 0);
 
     if (!cbw4 || !cbh4) return;
 
@@ -362,6 +364,7 @@ void dav1d_create_lf_mask_inter(Av1Filter *const lflvl,
     const int bh4 = imin(ih - by, b_dim[1]);
     const int bx4 = bx & 31;
     const int by4 = by & 31;
+    assert(bw4 >= 0 && bh4 >= 0);
 
     if (bw4 && bh4) {
         uint8_t (*level_cache_ptr)[4] = level_cache + by * b4_stride + bx;
@@ -385,6 +388,7 @@ void dav1d_create_lf_mask_inter(Av1Filter *const lflvl,
                           (b_dim[0] + ss_hor) >> ss_hor);
     const int cbh4 = imin(((ih + ss_ver) >> ss_ver) - (by >> ss_ver),
                           (b_dim[1] + ss_ver) >> ss_ver);
+    assert(cbw4 >= 0 && cbh4 >= 0);
 
     if (!cbw4 || !cbh4) return;
 
