@@ -89,15 +89,26 @@ void dav1d_default_picture_release(Dav1dPicture *const p, void *const cookie) {
 struct pic_ctx_context {
     Dav1dPicAllocator allocator;
     Dav1dPicture pic;
-    void *extra_ptr; /* MUST BE AT THE END */
+    Dav1dRef ref;
+    void *extra_data[];
 };
 
 static void free_buffer(const uint8_t *const data, void *const user_data) {
-    struct pic_ctx_context *pic_ctx = user_data;
+    Dav1dMemPoolBuffer *buf = (Dav1dMemPoolBuffer *)data;
+    struct pic_ctx_context *pic_ctx = buf->data;
 
     pic_ctx->allocator.release_picture_callback(&pic_ctx->pic,
                                                 pic_ctx->allocator.cookie);
-    free(pic_ctx);
+    dav1d_mem_pool_push(user_data, buf);
+}
+
+void dav1d_picture_free_itut_t35(const uint8_t *const data, void *const user_data) {
+    struct itut_t35_ctx_context *itut_t35_ctx = user_data;
+
+    for (size_t i = 0; i < itut_t35_ctx->n_itut_t35; i++)
+        dav1d_free(itut_t35_ctx->itut_t35[i].payload);
+    dav1d_free(itut_t35_ctx->itut_t35);
+    dav1d_free(itut_t35_ctx);
 }
 
 static int picture_alloc_with_edges(Dav1dContext *const c,
@@ -105,13 +116,10 @@ static int picture_alloc_with_edges(Dav1dContext *const c,
                                     const int w, const int h,
                                     Dav1dSequenceHeader *const seq_hdr, Dav1dRef *const seq_hdr_ref,
                                     Dav1dFrameHeader *const frame_hdr, Dav1dRef *const frame_hdr_ref,
-                                    Dav1dContentLightLevel *const content_light, Dav1dRef *const content_light_ref,
-                                    Dav1dMasteringDisplay *const mastering_display, Dav1dRef *const mastering_display_ref,
-                                    Dav1dITUTT35 *const itut_t35, Dav1dRef *const itut_t35_ref,
                                     const int bpc,
                                     const Dav1dDataProps *const props,
                                     Dav1dPicAllocator *const p_allocator,
-                                    const size_t extra, void **const extra_ptr)
+                                    void **const extra_ptr)
 {
     if (p->data[0]) {
         dav1d_log(c, "Picture already allocated!\n");
@@ -119,35 +127,30 @@ static int picture_alloc_with_edges(Dav1dContext *const c,
     }
     assert(bpc > 0 && bpc <= 16);
 
-    struct pic_ctx_context *pic_ctx = malloc(extra + sizeof(struct pic_ctx_context));
-    if (pic_ctx == NULL)
+    size_t extra = c->n_fc > 1 ? sizeof(atomic_int) * 2 : 0;
+    Dav1dMemPoolBuffer *buf = dav1d_mem_pool_pop(c->pic_ctx_pool,
+                                                 extra + sizeof(struct pic_ctx_context));
+    if (buf == NULL)
         return DAV1D_ERR(ENOMEM);
 
+    struct pic_ctx_context *pic_ctx = buf->data;
+
     p->p.w = w;
     p->p.h = h;
     p->seq_hdr = seq_hdr;
     p->frame_hdr = frame_hdr;
-    p->content_light = content_light;
-    p->mastering_display = mastering_display;
-    p->itut_t35 = itut_t35;
     p->p.layout = seq_hdr->layout;
     p->p.bpc = bpc;
     dav1d_data_props_set_defaults(&p->m);
     const int res = p_allocator->alloc_picture_callback(p, p_allocator->cookie);
     if (res < 0) {
-        free(pic_ctx);
+        dav1d_mem_pool_push(c->pic_ctx_pool, buf);
         return res;
     }
 
     pic_ctx->allocator = *p_allocator;
     pic_ctx->pic = *p;
-
-    if (!(p->ref = dav1d_ref_wrap(p->data[0], free_buffer, pic_ctx))) {
-        p_allocator->release_picture_callback(p, p_allocator->cookie);
-        free(pic_ctx);
-        dav1d_log(c, "Failed to wrap picture: %s\n", strerror(errno));
-        return DAV1D_ERR(ENOMEM);
-    }
+    p->ref = dav1d_ref_init(&pic_ctx->ref, buf, free_buffer, c->pic_ctx_pool, 0);
 
     p->seq_hdr_ref = seq_hdr_ref;
     if (seq_hdr_ref) dav1d_ref_inc(seq_hdr_ref);
@@ -155,50 +158,70 @@ static int picture_alloc_with_edges(Dav1dContext *const c,
     p->frame_hdr_ref = frame_hdr_ref;
     if (frame_hdr_ref) dav1d_ref_inc(frame_hdr_ref);
 
-    dav1d_data_props_copy(&p->m, props);
-
     if (extra && extra_ptr)
-        *extra_ptr = &pic_ctx->extra_ptr;
-
-    p->content_light_ref = content_light_ref;
-    if (content_light_ref) dav1d_ref_inc(content_light_ref);
-
-    p->mastering_display_ref = mastering_display_ref;
-    if (mastering_display_ref) dav1d_ref_inc(mastering_display_ref);
-
-    p->itut_t35_ref = itut_t35_ref;
-    if (itut_t35_ref) dav1d_ref_inc(itut_t35_ref);
+        *extra_ptr = &pic_ctx->extra_data;
 
     return 0;
 }
 
+void dav1d_picture_copy_props(Dav1dPicture *const p,
+                              Dav1dContentLightLevel *const content_light, Dav1dRef *const content_light_ref,
+                              Dav1dMasteringDisplay *const mastering_display, Dav1dRef *const mastering_display_ref,
+                              Dav1dITUTT35 *const itut_t35, Dav1dRef *itut_t35_ref, size_t n_itut_t35,
+                              const Dav1dDataProps *const props)
+{
+    dav1d_data_props_copy(&p->m, props);
+
+    dav1d_ref_dec(&p->content_light_ref);
+    p->content_light_ref = content_light_ref;
+    p->content_light = content_light;
+    if (content_light_ref) dav1d_ref_inc(content_light_ref);
+
+    dav1d_ref_dec(&p->mastering_display_ref);
+    p->mastering_display_ref = mastering_display_ref;
+    p->mastering_display = mastering_display;
+    if (mastering_display_ref) dav1d_ref_inc(mastering_display_ref);
+
+    dav1d_ref_dec(&p->itut_t35_ref);
+    p->itut_t35_ref = itut_t35_ref;
+    p->itut_t35 = itut_t35;
+    p->n_itut_t35 = n_itut_t35;
+    if (itut_t35_ref) dav1d_ref_inc(itut_t35_ref);
+}
+
 int dav1d_thread_picture_alloc(Dav1dContext *const c, Dav1dFrameContext *const f,
                                const int bpc)
 {
     Dav1dThreadPicture *const p = &f->sr_cur;
-    const int have_frame_mt = c->n_fc > 1;
 
     const int res =
         picture_alloc_with_edges(c, &p->p, f->frame_hdr->width[1], f->frame_hdr->height,
                                  f->seq_hdr, f->seq_hdr_ref,
                                  f->frame_hdr, f->frame_hdr_ref,
-                                 c->content_light, c->content_light_ref,
-                                 c->mastering_display, c->mastering_display_ref,
-                                 c->itut_t35, c->itut_t35_ref,
                                  bpc, &f->tile[0].data.m, &c->allocator,
-                                 have_frame_mt ? sizeof(atomic_int) * 2 : 0,
                                  (void **) &p->progress);
     if (res) return res;
 
+    dav1d_picture_copy_props(&p->p, c->content_light, c->content_light_ref,
+                             c->mastering_display, c->mastering_display_ref,
+                             c->itut_t35, c->itut_t35_ref, c->n_itut_t35,
+                             &f->tile[0].data.m);
+
     // Must be removed from the context after being attached to the frame
     dav1d_ref_dec(&c->itut_t35_ref);
     c->itut_t35 = NULL;
+    c->n_itut_t35 = 0;
 
+    // Don't clear these flags from c->frame_flags if the frame is not visible.
+    // This way they will be added to the next visible frame too.
+    const int flags_mask = (f->frame_hdr->show_frame || c->output_invisible_frames)
+                           ? 0 : (PICTURE_FLAG_NEW_SEQUENCE | PICTURE_FLAG_NEW_OP_PARAMS_INFO);
     p->flags = c->frame_flags;
-    c->frame_flags = 0;
+    c->frame_flags &= flags_mask;
 
     p->visible = f->frame_hdr->show_frame;
-    if (have_frame_mt) {
+    p->showable = f->frame_hdr->showable_frame;
+    if (c->n_fc > 1) {
         atomic_init(&p->progress[0], 0);
         atomic_init(&p->progress[1], 0);
     }
@@ -208,43 +231,48 @@ int dav1d_thread_picture_alloc(Dav1dContext *const c, Dav1dFrameContext *const f
 int dav1d_picture_alloc_copy(Dav1dContext *const c, Dav1dPicture *const dst, const int w,
                              const Dav1dPicture *const src)
 {
-    struct pic_ctx_context *const pic_ctx = src->ref->user_data;
+    Dav1dMemPoolBuffer *const buf = (Dav1dMemPoolBuffer *)src->ref->const_data;
+    struct pic_ctx_context *const pic_ctx = buf->data;
     const int res = picture_alloc_with_edges(c, dst, w, src->p.h,
                                              src->seq_hdr, src->seq_hdr_ref,
                                              src->frame_hdr, src->frame_hdr_ref,
-                                             src->content_light, src->content_light_ref,
-                                             src->mastering_display, src->mastering_display_ref,
-                                             src->itut_t35, src->itut_t35_ref,
                                              src->p.bpc, &src->m, &pic_ctx->allocator,
-                                             0, NULL);
-    return res;
+                                             NULL);
+    if (res) return res;
+
+    dav1d_picture_copy_props(dst, src->content_light, src->content_light_ref,
+                             src->mastering_display, src->mastering_display_ref,
+                             src->itut_t35, src->itut_t35_ref, src->n_itut_t35,
+                             &src->m);
+
+    return 0;
 }
 
 void dav1d_picture_ref(Dav1dPicture *const dst, const Dav1dPicture *const src) {
-    validate_input(dst != NULL);
-    validate_input(dst->data[0] == NULL);
-    validate_input(src != NULL);
+    assert(dst != NULL);
+    assert(dst->data[0] == NULL);
+    assert(src != NULL);
 
     if (src->ref) {
-        validate_input(src->data[0] != NULL);
+        assert(src->data[0] != NULL);
         dav1d_ref_inc(src->ref);
-        if (src->frame_hdr_ref) dav1d_ref_inc(src->frame_hdr_ref);
-        if (src->seq_hdr_ref) dav1d_ref_inc(src->seq_hdr_ref);
-        if (src->m.user_data.ref) dav1d_ref_inc(src->m.user_data.ref);
-        if (src->content_light_ref) dav1d_ref_inc(src->content_light_ref);
-        if (src->mastering_display_ref) dav1d_ref_inc(src->mastering_display_ref);
-        if (src->itut_t35_ref) dav1d_ref_inc(src->itut_t35_ref);
     }
+    if (src->frame_hdr_ref) dav1d_ref_inc(src->frame_hdr_ref);
+    if (src->seq_hdr_ref) dav1d_ref_inc(src->seq_hdr_ref);
+    if (src->m.user_data.ref) dav1d_ref_inc(src->m.user_data.ref);
+    if (src->content_light_ref) dav1d_ref_inc(src->content_light_ref);
+    if (src->mastering_display_ref) dav1d_ref_inc(src->mastering_display_ref);
+    if (src->itut_t35_ref) dav1d_ref_inc(src->itut_t35_ref);
     *dst = *src;
 }
 
 void dav1d_picture_move_ref(Dav1dPicture *const dst, Dav1dPicture *const src) {
-    validate_input(dst != NULL);
-    validate_input(dst->data[0] == NULL);
-    validate_input(src != NULL);
+    assert(dst != NULL);
+    assert(dst->data[0] == NULL);
+    assert(src != NULL);
 
     if (src->ref)
-        validate_input(src->data[0] != NULL);
+        assert(src->data[0] != NULL);
 
     *dst = *src;
     memset(src, 0, sizeof(*src));
@@ -255,6 +283,7 @@ void dav1d_thread_picture_ref(Dav1dThreadPicture *const dst,
 {
     dav1d_picture_ref(&dst->p, &src->p);
     dst->visible = src->visible;
+    dst->showable = src->showable;
     dst->progress = src->progress;
     dst->flags = src->flags;
 }
@@ -264,6 +293,7 @@ void dav1d_thread_picture_move_ref(Dav1dThreadPicture *const dst,
 {
     dav1d_picture_move_ref(&dst->p, &src->p);
     dst->visible = src->visible;
+    dst->showable = src->showable;
     dst->progress = src->progress;
     dst->flags = src->flags;
     memset(src, 0, sizeof(*src));
@@ -275,13 +305,13 @@ void dav1d_picture_unref_internal(Dav1dPicture *const p) {
     if (p->ref) {
         validate_input(p->data[0] != NULL);
         dav1d_ref_dec(&p->ref);
-        dav1d_ref_dec(&p->seq_hdr_ref);
-        dav1d_ref_dec(&p->frame_hdr_ref);
-        dav1d_ref_dec(&p->m.user_data.ref);
-        dav1d_ref_dec(&p->content_light_ref);
-        dav1d_ref_dec(&p->mastering_display_ref);
-        dav1d_ref_dec(&p->itut_t35_ref);
     }
+    dav1d_ref_dec(&p->seq_hdr_ref);
+    dav1d_ref_dec(&p->frame_hdr_ref);
+    dav1d_ref_dec(&p->m.user_data.ref);
+    dav1d_ref_dec(&p->content_light_ref);
+    dav1d_ref_dec(&p->mastering_display_ref);
+    dav1d_ref_dec(&p->itut_t35_ref);
     memset(p, 0, sizeof(*p));
     dav1d_data_props_set_defaults(&p->m);
 }
